@@ -55,6 +55,7 @@
 #include <osmocom/sgsn/gprs_ranap.h>
 #include <osmocom/sgsn/gprs_gmm_fsm.h>
 #include <osmocom/sgsn/gprs_mm_state_gb_fsm.h>
+#include <osmocom/sgsn/gprs_mm_state_iu_fsm.h>
 #include <osmocom/sgsn/gtp_ggsn.h>
 #include <osmocom/sgsn/gtp_mme.h>
 #include <osmocom/sgsn/sgsn_rim.h>
@@ -142,7 +143,7 @@ struct sgsn_pdp_ctx *sgsn_create_pdp_ctx(struct sgsn_ggsn_ctx *ggsn,
 					 uint16_t nsapi,
 					 struct tlv_parsed *tp)
 {
-	struct gprs_ra_id raid;
+	struct osmo_routing_area_id raid = {};
 	struct sgsn_pdp_ctx *pctx;
 	struct pdp_t *pdp;
 	uint64_t imsi_ui64;
@@ -166,6 +167,9 @@ struct sgsn_pdp_ctx *sgsn_create_pdp_ctx(struct sgsn_ggsn_ctx *ggsn,
 	pdp->priv = pctx;
 	pctx->lib = pdp;
 
+	/* Back up our own local TEID in case we update the library one with RNC TEID when setting up Direct Tunnel: */
+	pctx->sgsn_teid_own = pdp->teid_own;
+
 	//pdp->peer =	/* sockaddr_in of GGSN (receive) */
 	//pdp->ipif =	/* not used by library */
 	pdp->version = ggsn->gtp_version;
@@ -269,9 +273,9 @@ struct sgsn_pdp_ctx *sgsn_create_pdp_ctx(struct sgsn_ggsn_ctx *ggsn,
 	/* Routing Area Identifier with LAC and RAC fixed values, as
 	 * requested in 29.006 7.3.1 */
 	raid = mmctx->ra;
-	raid.lac = 0xFFFE;
+	raid.lac.lac = 0xFFFE;
 	raid.rac = 0xFF;
-	gsm48_encode_ra((struct gsm48_ra_id *)pdp->rai.v, &raid);
+	osmo_routing_area_id_encode_buf(pdp->rai.v, pdp->rai.l, &raid);
 
 	/* Encode User Location Information accordint to TS 29.060 7.7.51 */
 	pdp->userloc_given = 1;
@@ -286,20 +290,23 @@ struct sgsn_pdp_ctx *sgsn_create_pdp_ctx(struct sgsn_ggsn_ctx *ggsn,
 		pdp->userloc_given = 1;
 		pdp->userloc.l = 8;
 		pdp->userloc.v[0] = 0; /* CGI for GERAN */
-		bssgp_create_cell_id(&pdp->userloc.v[1], &mmctx->ra, mmctx->gb.cell_id);
+		bssgp_create_cell_id2(&pdp->userloc.v[1], 8, &mmctx->ra, mmctx->gb.cell_id);
 		break;
 	case MM_CTX_T_UTRAN_Iu:
 		pdp->userloc.v[0] = 1; /* SAI for UTRAN */
 		/* SAI is like CGI but with SAC instead of CID, so we can abuse this function */
-		bssgp_create_cell_id(&pdp->userloc.v[1], &mmctx->ra, mmctx->iu.sac);
+		bssgp_create_cell_id2(&pdp->userloc.v[1], 8, &mmctx->ra, mmctx->iu.sac);
 		break;
 	}
 
-	/* include the IMEI(SV) */
-	pdp->imeisv_given = 1;
-	gsm48_encode_bcd_number(&pdp->imeisv.v[0], 8, 0, mmctx->imei);
-	pdp->imeisv.l = pdp->imeisv.v[0];
-	memmove(&pdp->imeisv.v[0], &pdp->imeisv.v[1], 8);
+	/* optional include the IMEI(SV) */
+	if (mmctx->imei[0] != '\0') {
+		memset(&pdp->imeisv.v[0], 0, 8);
+		pdp->imeisv_given = 1;
+		gsm48_encode_bcd_number(&pdp->imeisv.v[0], 8, 0, mmctx->imei);
+		pdp->imeisv.l = 8;
+		memmove(&pdp->imeisv.v[0], &pdp->imeisv.v[1], 8);
+	}
 
 	/* change pdp state to 'requested' */
 	pctx->state = PDP_STATE_CR_REQ;
@@ -465,13 +472,6 @@ reject:
 	return EOF;
 }
 
-void sgsn_pdp_upd_gtp_u(struct sgsn_pdp_ctx *pdp, void *addr, size_t alen)
-{
-	pdp->lib->gsnlu.l = alen;
-	memcpy(pdp->lib->gsnlu.v, addr, alen);
-	gtp_update_context(pdp->ggsn->gsn, pdp->lib, pdp, &pdp->lib->hisaddr0);
-}
-
 void sgsn_ggsn_echo_req(struct sgsn_ggsn_ctx *ggc)
 {
 	LOGGGSN(ggc, LOGL_INFO, "GTP Tx Echo Request\n");
@@ -652,6 +652,52 @@ static int cb_conf(int type, int cause, struct pdp_t *pdp, void *cbp)
 	return 0;
 }
 
+/* Called whenever a PDP context is updated from the GGSN for any reason */
+static int cb_update_context_ind(struct pdp_t *pdp)
+{
+	struct sgsn_pdp_ctx *pctx;
+	struct sgsn_mm_ctx *mm;
+	int rc;
+
+	LOGPDPX(DGPRS, LOGL_INFO, pdp, "Context %p was updated\n", pdp);
+
+	pctx = pdp->priv;
+	if (!pctx) {
+		LOGP(DGPRS, LOGL_NOTICE,
+		     "GTP DATA IND from GGSN for unknown PDP\n");
+		return -EIO;
+	}
+	mm = pctx->mm;
+	if (!mm) {
+		LOGP(DGPRS, LOGL_ERROR,
+		     "PDP context (address=%u) without MM context!\n",
+		     pctx->address);
+		return -EIO;
+	}
+
+	if (mm->ran_type == MM_CTX_T_UTRAN_Iu) {
+#ifdef BUILD_IU
+		if (pdp->dir_tun_flags.v[0] & 0x04) { /* EI bit set ? */
+			/* GGSN informed us that it received an Error Indication when sending DL data to the RNC.
+			 * This probably means the RNC lost its state, aka crashed or was rebooted.
+			 * Page the UE so it re-creates the state at the RNC. */
+			LOGMMCTXP(LOGL_INFO, mm,
+				  "GGSN received ErrorInd from RNC while tx DL data. Paging UE in state %s\n",
+				  osmo_fsm_inst_state_name(mm->gmm_fsm));
+			rc = osmo_fsm_inst_dispatch(mm->iu.mm_state_fsm, E_PMM_RX_GGSN_GTPU_DT_EI, pctx);
+			rc = gtp_update_context_resp(sgsn->gsn, pdp,
+				 GTPCAUSE_ACC_REQ);
+			ranap_iu_page_ps(mm->imsi, &mm->p_tmsi, mm->ra.lac.lac, mm->ra.rac);
+			return rc;
+		}
+#endif
+	}
+
+	rc = gtp_update_context_resp(sgsn->gsn, pdp,
+				 GTPCAUSE_ACC_REQ);
+	return rc;
+}
+
 /* Called whenever a PDP context is deleted for any reason */
 static int cb_delete_context(struct pdp_t *pdp)
 {
@@ -783,7 +829,9 @@ static int cb_data_ind(struct pdp_t *lib, void *packet, unsigned int len)
 #ifdef BUILD_IU
 		/* Ignore the packet for now and page the UE to get the RAB
 		 * reestablished */
-		ranap_iu_page_ps(mm->imsi, &mm->p_tmsi, mm->ra.lac, mm->ra.rac);
+		LOGMMCTXP(LOGL_INFO, mm, "Rx GTP for UE in PMM state %s, paging it\n",
+			  osmo_fsm_inst_state_name(mm->iu.mm_state_fsm));
+		ranap_iu_page_ps(mm->imsi, &mm->p_tmsi, mm->ra.lac.lac, mm->ra.rac);
 
 		return 0;
 #else
@@ -845,7 +893,7 @@ static int cb_data_ind(struct pdp_t *lib, void *packet, unsigned int len)
 }
 
 /* Called by SNDCP when it has received/re-assembled a N-PDU */
-int sgsn_gtp_data_req(struct gprs_ra_id *ra_id, int32_t tlli, uint8_t nsapi,
+int sgsn_gtp_data_req(struct osmo_routing_area_id *ra_id, int32_t tlli, uint8_t nsapi,
 			 struct msgb *msg, uint32_t npdu_len, uint8_t *npdu)
 {
 	struct sgsn_mm_ctx *mmctx;
@@ -946,6 +994,7 @@ int sgsn_gtp_init(struct sgsn_instance *sgi)
 	}
 
 	/* Register callbackcs with libgtp */
+	gtp_set_cb_update_context_ind(gsn, cb_update_context_ind);
 	gtp_set_cb_delete_context(gsn, cb_delete_context);
 	gtp_set_cb_conf(gsn, cb_conf);
 	gtp_set_cb_recovery3(gsn, cb_recovery3);
